import random
import string


class TestUtilsMixin(object):
    """Mixin that provides extra testing utilities."""

    def assertNotRaises(self, func, *args, **kwargs):
        """Ensure that the an exception isn't raised.

        An AssertionError is raised if calling func with
        the given argument triggers any exception.
        """
        try:
            return func(*args, **kwargs)
        except Exception as e:
            msg = 'Expected no exception, but got {.__name__}'
            raise AssertionError(msg.format(e.__class__))


def random_string(size):
    """
    Generate a random string of size `size`.
    """
    choices = string.lowercase + string.uppercase + string.digits + '\n'
    return ''.join(random.choice(choices) for _ in range(size))



import array
import functools
import io
import os
import unittest

from heatshrink.streams import EncodedFile

from .constants import TEXT, COMPRESSED
from .utils import TestUtilsMixin, random_string

TEST_FILENAME = 'test_{}_tmp'.format(os.getpid())


class EncodedFileTest(TestUtilsMixin, unittest.TestCase):
    def setUp(self):
        self.fp = EncodedFile(TEST_FILENAME, 'wb')

    def tearDown(self):
        if self.fp:
            self.fp.close()

        # Cleanup temporary file
        if os.path.exists(TEST_FILENAME):
            os.unlink(TEST_FILENAME)

    def test_bad_args(self):
        self.assertRaises(TypeError, EncodedFile, None)
        self.assertRaises(ValueError, EncodedFile, None, mode='eb')
        fp = self.assertNotRaises(EncodedFile, TEST_FILENAME, invalid_param=True)
        fp.close()

    def test_open_missing_file(self):
        self.assertRaises(IOError, EncodedFile, 'does_not_exist.txt')

    def test_mode_attribute_is_readonly(self):
        self.assertEqual(self.fp.mode, 'wb')
        with self.assertRaises(AttributeError):
            self.fp.mode = 'rb'

    def test_different_compress_params(self):
        # List of a tuple containing (window_sz2, lookahead_sz2)
        encode_params = [
            (8, 3),
            (11, 6),
            (4, 3),
            (15, 9),
        ]

        encoded = []
        for (window_sz2, lookahead_sz2) in encode_params:
            kwargs = {
                'window_sz2': window_sz2,
                'lookahead_sz2': lookahead_sz2
            }
            with io.BytesIO() as dst:

                with EncodedFile(dst, 'wb', **kwargs) as fp:
                    fp.write(TEXT)

                encoded.append(dst.getvalue())

        # Ensure that all have different values
        self.assertEqual(len(encoded), len(set(encoded)))

    def test_invalid_modes(self):
        data = io.BytesIO()

        for mode in ['a+', 'w+', 'ab', 'r+', 'U', 'x', 'xb']:
            with self.assertRaisesRegexp(ValueError, '^Invalid mode: .*$'):
                EncodedFile(data, mode=mode)

    def test_round_trip(self):
        write_str = b'Testing\nAnd Stuff'

        self.fp.write(write_str)
        self.fp.close()

        self.fp = EncodedFile(TEST_FILENAME)
        self.assertEqual(self.fp.read(), write_str)

    def test_with_large_files(self):
        test_sizes = [1000, 10000, 100000]

        for size in test_sizes:
            contents = random_string(size)

            with EncodedFile(TEST_FILENAME, mode='wb') as fp:
                fp.write(contents)

            with EncodedFile(TEST_FILENAME) as fp:
                read_str = fp.read()

            if read_str != contents:
                msg = ('Decoded and original file contents '
                       'do not match for size: {}')
                self.fail(msg.format(size))

    def test_buffered_large_files(self):
        test_sizes = [1000, 10000, 100000]

        for size in test_sizes:
            contents = random_string(size)

            with EncodedFile(TEST_FILENAME, mode='wb') as fp:
                fp.write(contents)

            with EncodedFile(TEST_FILENAME) as fp:
                # Read small buffer sizes
                read_func = functools.partial(fp.read, 512)
                read_str = ''.join([s for s in iter(read_func, '')])

            if read_str != contents:
                msg = ('Decoded and original file contents '
                       'do not match for size: {}')
                self.fail(msg.format(size))

    def test_with_file_object(self):
        plain_file = open(TEST_FILENAME, 'wb')

        with EncodedFile(plain_file, mode='wb') as encoded_file:
            encoded_file.write(TEXT)

        self.assertTrue(encoded_file.closed)
        # Shouldn't close the file, as it doesn't own it
        self.assertFalse(plain_file.closed)
        plain_file.close()

        with open(TEST_FILENAME, 'rb') as fp:
            self.assertTrue(len(fp.read()) > 0)

    def test_closed_true_when_file_closed(self):
        self.assertFalse(self.fp.closed)
        self.fp.close()
        self.assertTrue(self.fp.closed)

    def test_context_manager(self):
        with EncodedFile(TEST_FILENAME, mode='wb') as fp:
            fp.write(b'Testing\n')
            fp.write(b'One, two...')

        self.assertTrue(fp.closed)

    def test_operations_on_closed_file(self):
        self.fp.close()
        self.assertRaises(ValueError, self.fp.write, b'abcde')
        self.assertRaises(ValueError, self.fp.seek, 0)

        self.fp = EncodedFile(TEST_FILENAME, 'rb')
        self.fp.close()
        self.assertRaises(ValueError, self.fp.read)
        self.assertRaises(ValueError, self.fp.seek, 0)

    def test_cannot_write_in_read_mode(self):
        # Write some junk data
        self.fp.write(b'abcde')
        self.fp.close()

        self.fp = EncodedFile(TEST_FILENAME)
        self.assertTrue(self.fp.readable())
        self.assertFalse(self.fp.writable())
        self.assertRaises(IOError, self.fp.write, b'abcde')

    def test_cannot_read_in_write_mode(self):
        self.assertTrue(self.fp.writable())
        self.assertFalse(self.fp.readable())
        self.assertRaises(IOError, self.fp.read)

    #################
    # Seeking
    #################
    def test_seeking_forwards(self):
        contents = TEXT

        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            self.assertEqual(fp.read(100), contents[:100])
            fp.seek(150)  # Move 50 forwards
            self.assertEqual(fp.read(100), contents[150:250])

    def test_seeking_backwards(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            contents = fp.read(100)
            fp.seek(0)
            self.assertEqual(fp.read(100), contents)

    def test_seeking_forward_from_current(self):
        contents = TEXT

        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            self.assertEqual(fp.read(100), contents[:100])
            fp.seek(50, io.SEEK_CUR)  # Move 50 forwards
            self.assertEqual(fp.read(100), contents[150:250])

    def test_seeking_backwards_from_current(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            contents = fp.read()
            fp.seek(-100, io.SEEK_CUR)
            self.assertEqual(fp.read(), contents[-100:])

    def test_seeking_beyond_beginning_from_current(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            self.assertRaises(IOError, fp.seek, -100, io.SEEK_CUR)

    def test_seeking_from_end(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            self.assertEqual(fp.read(100), TEXT[:100])
            seeked_pos = fp.seek(-100, io.SEEK_END)
            self.assertEqual(seeked_pos, len(TEXT) - 100)
            self.assertEqual(fp.read(100), TEXT[-100:])

    def test_seeking_from_end_beyond_beginning(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            # Go to end to get size
            size = fp.seek(0, io.SEEK_END)
            # Go to beginning
            self.assertNotRaises(fp.seek, -size, io.SEEK_END)
            # One before beginning
            self.assertRaises(IOError, fp.seek, -size - 1, io.SEEK_END)

    def test_tell(self):
        with io.BytesIO() as dst:
            with EncodedFile(dst, mode='wb') as fp:
                bytes_written = fp.write(b'abcde')
                self.assertEqual(fp.tell(), bytes_written)

            dst.seek(0)  # Reset

            with EncodedFile(dst) as fp:
                fp.read(3)
                self.assertEqual(fp.tell(), 3)

    def test_peek(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            pdata = fp.peek()
            self.assertNotEqual(len(pdata), 0)
            self.assertTrue(TEXT.startswith(pdata))
            self.assertEqual(fp.read(), TEXT)

    #################
    # Reading
    #################
    def test_read_whole_file(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            self.assertEqual(fp.read(), TEXT)

    def test_read_buffered(self):
        READ_SIZE = 128
        offset = 0

        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            read_buf = functools.partial(fp.read, READ_SIZE)

            for i, contents in enumerate(iter(read_buf, '')):
                offset = READ_SIZE * i
                self.assertEqual(
                    contents,
                    TEXT[offset: offset + READ_SIZE]
                )

    def test_read_one_char(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            for c in TEXT:
                self.assertEqual(fp.read(1), c)

    def test_read1(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            blocks = [buf for buf in iter(fp.read1, '')]
            self.assertEqual(b''.join(blocks), TEXT)
            self.assertEqual(fp.read1(), b'')

    def test_read1_0(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            self.assertEqual(fp.read1(0), b'')

    def test_readinto(self):
        with io.BytesIO() as dst:
            with EncodedFile(dst, mode='wb') as fp:
                fp.write(b'abcde')

            dst.seek(0)  # Reset

            with EncodedFile(dst) as fp:
                a = array.array('b', b'x' * 10)  # Fill with junk
                n = fp.readinto(a)
                try:
                    # Python 3
                    self.assertEqual(b'abcde', a.tobytes()[:n])
                except AttributeError:
                    self.assertEqual(b'abcde', a.tostring()[:n])

    def test_readline(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            lines = TEXT.splitlines()

            # Could also use zip
            for i, line in enumerate(iter(fp.readline, '')):
                self.assertEqual(line, lines[i] + '\n')

    def test_readline_iterator(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            lines = TEXT.splitlines()

            for file_line, original_line in zip(fp, lines):
                self.assertEqual(file_line, original_line + '\n')

    def test_readlines(self):
        with EncodedFile(io.BytesIO(COMPRESSED)) as fp:
            lines = fp.readlines()
            self.assertEqual(''.join(lines), TEXT)

    #################
    # Writing
    #################
    def test_write_buffered(self):
        BUFFER_SIZE = 16
        # BytesIO makes it easy to buffer
        text_buf = io.BytesIO(TEXT.encode('utf8'))

        with io.BytesIO() as dst:
            with EncodedFile(dst, mode='wb') as fp:
                while True:
                    chunk = text_buf.read(BUFFER_SIZE)

                    if not chunk:
                        break

                    fp.write(chunk)

            self.assertEqual(dst.getvalue(), COMPRESSED)

    def test_remaining_data_flushed_on_close(self):
        with io.BytesIO() as dst:
            fp = EncodedFile(dst, mode='wb')
            fp.write(TEXT)
            # Not flusshed
            self.assertEqual(len(dst.getvalue()), 0)
            # Flush
            fp.close()
            self.assertTrue(len(dst.getvalue()) > 0)

    def test_writelines(self):
        with io.BytesIO(TEXT) as fp:
            lines = fp.readlines()

        with io.BytesIO() as dst:
            with EncodedFile(dst, mode='wb') as fp:
                fp.writelines(lines)

            self.assertEqual(dst.getvalue(), COMPRESSED)



from libc.stdint cimport uint8_t, uint16_t


cdef extern from "_heatshrink/heatshrink_common.h":
    cdef int HEATSHRINK_MIN_WINDOW_BITS
    cdef int HEATSHRINK_MAX_WINDOW_BITS

    cdef int HEATSHRINK_MIN_LOOKAHEAD_BITS


cdef extern from "_heatshrink/heatshrink_encoder.h":
    ctypedef struct heatshrink_encoder:
        uint8_t window_sz2
        uint8_t lookahead_sz2

    ctypedef enum HSE_sink_res:
        HSER_SINK_OK,
        HSER_SINK_ERROR_NULL = -1
        HSER_SINK_ERROR_MISUSE = -1

    ctypedef enum HSE_poll_res:
        HSER_POLL_EMPTY
        HSER_POLL_MORE
        HSER_POLL_ERROR_NULL = -1
        HSER_POLL_ERROR_MISUSE = -2

    ctypedef enum HSE_finish_res:
        HSER_FINISH_DONE
        HSER_FINISH_MORE
        HSER_FINISH_ERROR_NULL = -1

    heatshrink_encoder *heatshrink_encoder_alloc(uint8_t window_sz2, uint8_t lookahead_sz2)

    void heatshrink_encoder_free(heatshrink_encoder *hse)

    HSE_sink_res heatshrink_encoder_sink(heatshrink_encoder *hse, uint8_t *in_buf,
                                         size_t size, size_t *input_size) nogil

    HSE_poll_res heatshrink_encoder_poll(heatshrink_encoder *hse, uint8_t *out_buf,
                                         size_t out_buf_size, size_t *output_size) nogil

    HSE_finish_res heatshrink_encoder_finish(heatshrink_encoder *hse)


cdef extern from "_heatshrink/heatshrink_decoder.h":
    ctypedef struct heatshrink_decoder:
        uint8_t window_sz2
        uint8_t lookahead_sz2

    ctypedef enum HSD_sink_res:
        HSDR_SINK_OK,
        HSDR_SINK_FULL
        HSDR_SINK_ERROR_NULL = -1

    ctypedef enum HSD_poll_res:
        HSDR_POLL_EMPTY
        HSDR_POLL_MORE
        HSDR_POLL_ERROR_NULL = -1
        HSDR_POLL_ERROR_UNKNOWN = -2

    ctypedef enum HSD_finish_res:
        HSDR_FINISH_DONE
        HSDR_FINISH_MORE
        HSDR_FINISH_ERROR_NULL = -1

    heatshrink_decoder *heatshrink_decoder_alloc(uint16_t input_buffer_size,
                                                 uint8_t expansion_buffer_sz2,
                                                 uint8_t lookahead_sz2)

    void heatshrink_decoder_free(heatshrink_decoder *hsd)

    HSD_sink_res heatshrink_decoder_sink(heatshrink_decoder *hsd, uint8_t *in_buf,
                                         size_t size, size_t *input_size) nogil

    HSD_poll_res heatshrink_decoder_poll(heatshrink_decoder *hsd, uint8_t *out_buf,
                                         size_t out_buf_size, size_t *output_size) nogil

    HSD_finish_res heatshrink_decoder_finish(heatshrink_decoder *hsd) nogil



from os import path
from codecs import open
from setuptools import setup, Extension

try:
    from Cython.Build import cythonize
except ImportError:
    USE_CYTHON = False
else:
    USE_CYTHON = True


# Compiled file extension to use. If we're not using Cython,
# just use the plain C file.
EXT = '.pyx' if USE_CYTHON else '.c'

heatshrink_module = Extension('heatshrink.core',
                              include_dirs=['.', './heatshrink/_heatshrink'],
                              extra_compile_args=['-std=c99'],
                              sources=['heatshrink/core' + EXT,
                                       'heatshrink/_heatshrink/heatshrink_encoder.c',
                                       'heatshrink/_heatshrink/heatshrink_decoder.c'])

if USE_CYTHON:
    extensions = cythonize([heatshrink_module])
else:
    extensions = [heatshrink_module]


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='Heatshrink',
      version='0.3.2',
      # Author details
      author='Antonis Kalou @ JOHAN Sports',
      author_email='antonis@johan-sports.com',
      # Project details
      description='Python bindings to the heatshrink library',
      long_description=long_description,
      url='https://github.com/johan-sports/pyheatshrink',
      license='ISC',

      classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: System :: Archiving :: Compression',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: ISC License (ISCL)',

        # Supported python versions
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
      ],

      keywords='compression binding heatshrink LZSS',
      test_suite='tests',
      packages=['heatshrink'],
      ext_modules=extensions)
